import re
from collections.abc import Iterator
from typing import Any

from citation_date import DOCKET_DATE_REGEX, decode_date
from citation_report import REPORT_REGEX, get_publisher_label
from pydantic import BaseModel, Field

from .misc import cull_extra, formerly, pp


class CitationConstructor(BaseModel):
    """Prefatorily, regex strings are defined so that a
    `re.Pattern` object can take advantage of the "group_name"
    assigned in the string.

    These are the docket styles with regex strings predefined:

    1. General Register
    2. Administrative Matter
    3. Administrative Case
    4. Bar Matter
    5. Office of the Court Administrator
    6. Presidential Electoral Tribunal
    7. Judicial Integrity Board
    8. Undocketed Case

    The CitationConstructor formalizes the assigned group names into
    their respective fields.

    Relatedly, it takes advantage of
    the `citation_date` and the `citation_report` libraries in
    generating the main `@pattern` since the regex strings above
    are only concerned with the `key` `num` `id` formula part
    of the docket, e.g. `GR` `No.` `123`... but not the accompanying
    date and report.
    """

    label: str = Field(
        ...,
        title="Docket Label",
        description="e.g. General Register, Administrative Matter",
    )
    short_category: str = Field(
        ..., title="Docket Category Shorthand", description="e.g. GR, AM"
    )
    group_name: str = Field(
        ...,
        title="Regex Group Name",
        description=(
            "e.g. 'gr_test_phrase' identifies that portion of the"
            "Match object that should be associated with the label."
        ),
    )
    init_name: str = Field(
        ...,
        title="Regex Initial Group Name",
        description="e.g. gr_mid, am_init; see .regexes for other group names",
    )
    docket_regex: str = Field(
        ...,
        title="Regex Expression Proper",
        description=(
            "The full regex expression which includes the groupnames referred to above."
        ),
    )
    key_regex: str = Field(
        ...,
        title="Regex Key",
        description="Regex portion to get the serial ids",
    )
    num_regex: str = Field(
        ...,
        title="Regex Num",
        description="Regex portion for the num keyword to get the serial ids",
    )

    @property
    def pattern(self) -> re.Pattern:
        """Construct the regex string and generate a full Pattern object from:

        1. `docket_regex`,
        2. `docket_date` defined in the citation-date library
        3. an optional `REPORT_REGEX` defined in the citation-report library

        Returns:
            Pattern: Combination of Docket and Report styles.
        """
        return re.compile(
            "".join(
                [
                    rf"{self.docket_regex}",
                    rf"(?P<extra_phrase>{formerly}?{pp}?){DOCKET_DATE_REGEX}",
                    rf"(?P<opt_report>\,\s*{REPORT_REGEX})?",
                ]
            ),
            re.I | re.X,
        )

    @property
    def key_num_pattern(self) -> re.Pattern:
        """Unlike full @pattern, this regex compiled object is limited to
        just the key and number elements, e.g. "GR No. 123" or "BP Blg. 45"
        """
        regex = rf"{self.key_regex}({self.num_regex})?"
        return re.compile(regex, re.I | re.X)

    def detect(self, raw: str) -> Iterator[dict[str, Any]]:
        """Logic: if `self.init_name` Match group exists, get entire
        regex based on `self.group_name`, extract subgroups which will
        consist of `Docket` and `Report` parts.

        Args:
            raw (str): Text to evaluate

        Yields:
            Iterator[dict[str, Any]]: A dict that can fill up a Docket + Report pydantic BaseModel
        """  # noqa: E501
        for match in self.pattern.finditer(raw):
            if match.group(self.init_name):
                if ctx := match.group(self.group_name).strip(", "):
                    raw_id = cull_extra(self.key_num_pattern.sub("", ctx))
                    ids = raw_id.strip("()[] .,;")
                    raw_date = match.group("docket_date")
                    date_found = decode_date(raw_date, True)
                    if ids and date_found:
                        yield dict(
                            context=ctx,
                            short_category=self.short_category,
                            category=self.label,
                            ids=ids,
                            docket_date=date_found,
                            publisher=get_publisher_label(match),
                            volpubpage=match.group("volpubpage"),
                            volume=match.group("volume"),
                            page=match.group("page"),
                        )



from .constructed_ac import (
    CitationAC,
    ac_key,
    ac_phrases,
    constructed_ac,
)
from .constructed_am import (
    CitationAM,
    am_key,
    am_phrases,
    constructed_am,
)
from .constructed_bm import (
    CitationBM,
    bm_key,
    bm_phrases,
    constructed_bm,
)
from .constructed_gr import (
    CitationGR,
    constructed_gr,
    gr_key,
    gr_phrases,
    l_key,
)
from .constructed_jib import (
    CitationJIB,
    constructed_jib,
    jib_key,
    jib_phrases,
)
from .constructed_oca import (
    CitationOCA,
    constructed_oca,
    oca_key,
    oca_phrases,
)
from .constructed_pet import (
    CitationPET,
    constructed_pet,
    pet_key,
    pet_phrases,
)
from .constructed_udk import (
    CitationUDK,
    constructed_udk,
    udk_key,
    udk_phrases,
)
from .models import (
    DOCKET_DATE_FORMAT,
    CitationConstructor,
    Docket,
    DocketCategory,
    DocketReportCitation,
    Num,
    cull_extra,
    formerly,
    gr_prefix_clean,
    is_statutory_rule,
    pp,
)

DocketReport = (
    CitationAC
    | CitationAM
    | CitationOCA
    | CitationBM
    | CitationGR
    | CitationPET
    | CitationJIB
    | CitationUDK
)



from collections.abc import Iterator
from typing import Self

from .models import CitationConstructor, DocketCategory, DocketReportCitation
from .models.misc import NUMBER_KEYWORD

separator = r"[\.\s]*"
digit = r"(\d{2}|P)[\d-]*"  # e.g. 323-23, 343-34

end_digit = r"""
    (\d{2,})| # AM OCA IPI No. P-07-2403, Feb. 06, 2008; see also OCA I.P.I. No. 17-4757
    RTJ| # AM OCA IPI No. 11-3800-RTJ, Jun. 19, 2017
    MTJ| # AM OCA IPI No. 04-1606-MTJ, Sep. 19, 2012
    P| # AM OCA IPI No. 09-3138-P, Oct. 22, 2013
    SB-J| # AM OCA IPI No. 10-25-SB-J, Jan. 15, 2013
    CA-J| # AM OCA IPI No. 11-184-CA-J, Jan. 31, 2012
    METC # AM OCA IPI NO. 06-11-392-METC, Jan. 15, 2007
"""

full_digit = rf"""
    {digit}
    ({end_digit})
"""

oca_key = rf"""
    (
        (
            a
            {separator}
            m
            {separator}
        )?
        (
            (OCA)|(O{separator}C{separator}A{separator})
        )
        \s*
        (
            (IPI)|(I{separator}P{separator}I{separator})
        )?
        \s*
        {NUMBER_KEYWORD}
    )
"""

required = rf"""
    (?P<oca_init>
        {oca_key}
    )
    (?P<oca_middle>
        {full_digit}
    )
    (?:
        {separator}
    )?
"""


oca_phrases = rf"""
    (?P<oca_phrase>
        {required}
        [\,\s]*
    )
"""

constructed_oca = CitationConstructor(
    label=DocketCategory.OCA.value,
    short_category=DocketCategory.OCA.name,
    group_name="oca_phrase",
    init_name="oca_init",
    docket_regex=oca_phrases,
    key_regex=oca_key,
    num_regex=NUMBER_KEYWORD,
)


class CitationOCA(DocketReportCitation):
    ...

    @classmethod
    def search(cls, text: str) -> Iterator[Self]:
        """Get all dockets matching the `OCA` docket pattern, inclusive of their optional Report object.

        Examples:
            >>> text1 = "AM OCA IPI No. P-07-2403, Feb. 06, 2008"
            >>> cite1 = next(CitationOCA.search(text1))
            >>> cite1.model_dump(exclude_none=True)
            {'context': 'AM OCA IPI No. P-07-2403', 'category': 'OCA', 'ids': 'P-07-2403', 'docket_date': datetime.date(2008, 2, 6)}
            >>> text2 = "OCA IPI No. 10-3450-P, Feb. 06, 2008"
            >>> cite2 = next(CitationOCA.search(text2))
            >>> cite2.model_dump(exclude_none=True)
            {'context': 'OCA IPI No. 10-3450-P', 'category': 'OCA', 'ids': '10-3450-P', 'docket_date': datetime.date(2008, 2, 6)}


        Args:
            text (str): Text to look for citation objects

        Yields:
            Iterator[Self]: Combination of Docket and Report pydantic model.
        """  # noqa E501
        for result in constructed_oca.detect(text):
            yield cls(**result)



from collections.abc import Iterator
from typing import Self

from .models import CitationConstructor, DocketCategory, DocketReportCitation, Num

separator = r"[,\.\s-]*"
digit = r"\d[\d-]*"  # e.g. 323-23, 343-34, L


gr_key = rf"""
    (
        (
            (?<!CA\s)  # CA GR
            (?<!CA-)  # CA-GR
            (?<!C\.A\.) # C.A.G.R.
            (?<!C\.A\.\s) # C.A. G.R.
            \b
            g
        )
        {separator}
        r
        {separator}
    )
"""

l_key = rf"""
    (
        \b # prevent capture of Angara v. Electoral Commission, 63 Phil. 139,
        (
            l|
            i # I is a mistype: De Ramos v. Court of Agrarian Relations, I-19555, May 29, 1964;
        )
        {separator} # Makes possible G.R. L-No. 5110  | # L. No. 464564
    )
"""  # noqa # E501


digits = rf"""
    (
        \b
        ({l_key})?
        {digit}
        \b
        \W* # possible comma
        (&\s*|and\s*)? # possible and between
    ){{1,5}}
    (
        {digit}
        \b
    )?
"""


gr_regular = rf"""
( # G.R. L-No. 5110, G.R. No. 5110, GR 100
    {gr_key}
    ({l_key})?
    ({Num.GR.allowed})?
    {digits}
)
"""

l_irregular = rf"""
( # L-No. 5110, L 5110, No
    {l_key}
    ({Num.GR.allowed})?
    {digits}
)
"""

n_irregular = rf"""
( # , Nos.
    \,\s+
    ({Num.GR.allowed})
    {digits}
)
"""

gr_phrases = rf"""
    (?P<gr_phrase>
        (
            (?P<gr_mid>
                (?P<gr_gist_r>{gr_regular})|
                (?P<gr_gist_l>{l_irregular})|
                (?P<gr_gist_n>{n_irregular})
            )
        ){{1,5}}
    )
"""

constructed_gr = CitationConstructor(
    label=DocketCategory.GR.value,
    short_category=DocketCategory.GR.name,
    group_name="gr_phrase",
    init_name="gr_mid",
    docket_regex=gr_phrases,
    key_regex=rf"{gr_key}({l_key})?",
    num_regex=Num.GR.allowed,
)


class CitationGR(DocketReportCitation):
    ...

    @classmethod
    def search(cls, text: str) -> Iterator[Self]:
        """Get all dockets matching the `GR` docket pattern, inclusive of their optional Report object.

        Examples:
            >>> text = "Bagong Alyansang Makabayan v. Zamora, G.R. Nos. 138570, 138572, 138587, 138680, 138698, October 10, 2000, 342 SCRA 449"
            >>> cite = next(CitationGR.search(text))
            >>> cite.model_dump(exclude_none=True)
            {'publisher': 'SCRA', 'volume': '342', 'page': '449', 'context': 'G.R. Nos. 138570, 138572, 138587, 138680, 138698', 'category': 'GR', 'ids': '138570, 138572, 138587, 138680, 138698', 'docket_date': datetime.date(2000, 10, 10)}

        Args:
            text (str): Text to look for citation objects

        Yields:
            Iterator[Self]: Combination of Docket and Report pydantic model.
        """  # noqa E501
        for result in constructed_gr.detect(text):
            yield cls(**result)



from collections.abc import Iterator
from typing import Self

from .models import CitationConstructor, DocketCategory, DocketReportCitation
from .models.misc import NUMBER_KEYWORD

separator = r"[\.\s]*"
full_digit = r"\d{4,}"  # e.g. 16915


udk_key = rf"""
    (
        u
        {separator}
        d
        {separator}
        k
        {separator}
        -?
        \s*
        ({NUMBER_KEYWORD})?
    )
"""  # UDK 16915. February 15, 2022, note the number symbol No. may be optional

required = rf"""
    (?P<udk_init>
        {udk_key}
    )
    (?P<udk_middle>
        {full_digit}
    )
    (?:
        {separator}
    )?
"""


udk_phrases = rf"""
    (?P<udk_phrase>
        {required}
        [\,\s]*
    )
"""

constructed_udk = CitationConstructor(
    label=DocketCategory.UDK.value,
    short_category=DocketCategory.UDK.name,
    group_name="udk_phrase",
    init_name="udk_init",
    docket_regex=udk_phrases,
    key_regex=udk_key,
    num_regex=NUMBER_KEYWORD,
)


class CitationUDK(DocketReportCitation):
    ...

    @classmethod
    def search(cls, text: str) -> Iterator[Self]:
        """Get all dockets matching the `UDK` docket pattern, inclusive of their optional Report object.

        Examples:
            >>> text1 = "UDK 16915, February 15, 2022"
            >>> cite = next(CitationUDK.search(text1))
            >>> cite.model_dump(exclude_none=True)
            {'context': 'UDK 16915', 'category': 'UDK', 'ids': '16915', 'docket_date': datetime.date(2022, 2, 15)}
            >>> text2 = "UDK-1111, February 15, 2021"
            >>> cite2 = next(CitationUDK.search(text2))
            >>> cite2.model_dump(exclude_none=True)
            {'context': 'UDK-1111', 'category': 'UDK', 'ids': '1111', 'docket_date': datetime.date(2021, 2, 15)}

        Args:
            text (str): Text to look for citation objects

        Yields:
            Iterator[Self]: Combination of Docket and Report pydantic model.
        """  # noqa E501
        for result in constructed_udk.detect(text):
            yield cls(**result)



from collections import namedtuple
from enum import Enum


class Point(namedtuple("Point", ["x", "y"])):
    """
    Based on two letters, get valid options and patterns
    """

    __slots__ = ()

    @property
    def permuted(self):
        return [
            "".join([self.x, self.y]),  # gr
            "".join([self.x, r"\.", self.y]),  # g.r
            "".join([self.x, self.y, r"\."]),  # gr.
            "".join([self.x, r"\.", self.y, r"\."]),  # g.r.
        ]


NUMBER_KEYWORD = r"""
    \b
    No # Number
    s? # the letter s, optional
    \.? # optional .
    \s*
"""


class Key(Enum):
    GR = Point("g", "r").permuted
    AM = Point("a", "m").permuted + [r"Adm\.\sMatter", r"Admin\sMatter"]
    BM = Point("b", "m").permuted + [r"Bar\sMatter"]
    AC = Point("a", "c").permuted + [
        r"A\.C\.\s\-\sCBD",
        r"A\.C\.\sCBD",
        r"Adm\.\sCase",
        r"Admin\sCase",
    ]


class Num(Enum):
    GR = Key.AM.value + Key.AC.value + Key.BM.value
    AM = Key.GR.value + Key.AC.value + Key.BM.value
    AC = Key.GR.value + Key.AM.value + Key.BM.value
    BM = Key.GR.value + Key.AM.value + Key.AC.value

    @property
    def allowed(self):
        """
        Prevent capture of "Nos" word when not preceded by the
        intended docket type. Get excluded raw literals combined
        as negative lookbehinds for keyword "Nos".
        """
        ex_list = [rf"(?<!{e}\s)" for e in self.value]
        pre_num = "".join(ex_list)
        return rf"{pre_num}{NUMBER_KEYWORD}"



from collections.abc import Iterator
from typing import Self

from .models import CitationConstructor, DocketCategory, DocketReportCitation, Num

separator = r"[,\.\s-]*"
two_digits = r"[\d-]{2,}"
l_digits = rf"\bL\-{two_digits}"
digits_alone = rf"\b\d{two_digits}"

acronyms = r"""
    \s?
    (
        P| # A.C. No. P-88-198, February 25, 1992, 206 SCRA 491.
        J| # Adm. Case No. 129-J, July 30, 1976, 72 SCRA 172.
        CBD| # A.C. No. CBD-174
        CFI| # Adm. Case No. 1701-CFI
        CJ|# Adm. Matter No. 584-CJ
        MJ|# ADM CASE No. 783-MJ
        SBC|#Adm. Case No. 545-SBC
        SB # A.C. No. SB-95-7-P
    )
    \s?
"""
letter = rf"""
    (
        \b
        {acronyms}
    )?
    [\d-]{{3,}} #  at least two digits and a dash
    ( # don't add \b  to capture "-Ret.""
        {acronyms}
    )?
"""

ac_key = rf"""
    (
        (
            a
            {separator}
            c
            {separator}
            (?:
                CBD # see A.C. CBD No. 190
                \s* # optional space
            )? # optional CBD,
        )|
        (
            \b
            adm
            (in)?
            (istrative)?
            {separator}
            (?:
                \b
                case
                \s* # optional space
            )?
        )
    )
"""

ac_num = rf"""
    (
        {ac_key}
        {Num.AC.allowed}
    )
"""

required = rf"""
    (?P<ac_init>
        {ac_num}
    )
    (?P<ac_middle>
        (
            ({letter})|
            ({l_digits})|
            ({digits_alone})
        )
    )
    (?:
        (
            [\,\s,\-\&]|
            and
        )*
    )?
"""

optional = rf"""
    (?P<ac_init_optional>
        {ac_num}
    )?
    (?P<ac_middle_optional>
        ({letter})|
        ({l_digits})|
        ({digits_alone})
    )?
    (?:
        (
            [\,\s,\-\&]|
            and
        )*
    )?
"""

ac_phrases = rf"""
    (?P<ac_phrase>
        ({required})
        ({optional}){{1,3}}
    )
"""


constructed_ac = CitationConstructor(
    label=DocketCategory.AC.value,
    short_category=DocketCategory.AC.name,
    group_name="ac_phrase",
    init_name="ac_init",
    docket_regex=ac_phrases,
    key_regex=ac_key,
    num_regex=Num.AC.allowed,
)


class CitationAC(DocketReportCitation):
    ...

    @classmethod
    def search(cls, text: str) -> Iterator[Self]:
        """Get all dockets matching the `AC` docket pattern, inclusive of their optional Report object.

        Examples:
            >>> text = "A.C. No. P-88-198, February 25, 1992, 206 SCRA 491."
            >>> cite = next(CitationAC.search(text))
            >>> cite.model_dump(exclude_none=True)
            {'publisher': 'SCRA', 'volume': '206', 'page': '491', 'context': 'A.C. No. P-88-198', 'category': 'AC', 'ids': 'P-88-198', 'docket_date': datetime.date(1992, 2, 25)}

        Args:
            text (str): Text to look for citation objects

        Yields:
            Iterator[Self]: Combination of Docket and Report pydantic model.
        """  # noqa E501
        for result in constructed_ac.detect(text):
            yield cls(**result)



import re

LEGACY_PREFIXED = r"""
    ^ # start
    (?:
        (?:
            No # shorthand for number
            s? # plural option
            \. # period
            \s+ # space/s
        )? # number
        (?:
            (?:
                L
                \s* # L-26353
                -? # L 12271
                \s* # L- 59592
            )|
            (?:
                I\s- # I -5458
            )|
            (?:
                I- # I-19555
            )|
            (?:
                I\.- # I.-12735
            )
        )
    )
    (?=\w+) # excluded alphanumeric character
"""

LEGACY_PREFIXED_LOOKALIKE = r"""
    ^ # start
    (?:
        (?:
            No # shorthand for number
            s? # plural option
            \. # period
            \s+ # space/s
        )? # number
        (?:
            L
            \s* # L-26353
            -? # L 12271
            \s* # L- 59592
        )|
        (?:
            I\s- # I -5458
        )|
        (?:
            I- # I-19555
        )|
        (?:
            I\.- # I.-12735
        )
    )
    [ILl] # necessary after the group
    \s?
"""


def remove_prefix_regex(regex_to_match: str, text: str):
    """Based on the `regex` passed, remove this from the start of the `text`"""
    match = re.search(regex_to_match, text, re.VERBOSE)
    if not match:
        return None
    return text.strip().removeprefix(match.group())


def replace_prefix_regex(regex_to_match: str, text: str, std: str):
    """Based on the `regex` passed, replace this from the start of the `text`
    with a standardized variant `std`."""
    match = re.search(regex_to_match, text, re.VERBOSE)
    if not match:
        return None
    return std + text.strip().removeprefix(match.group()).strip()


def gr_prefix_clean(text: str) -> str | None:
    """The GR (General Register) docket ID makes use of `L-xxx` as a prefix
    in some of its serialized ids.

    Since most legal documents are parsed via OCR, the translation is often
    errneous resulting in an L-`I`9863 instead of being L-`1`9863.

    This also deals with cases involving inconsistent formatting, e.g.
    `No. L-12414`.

    If the regex patterns find the inconsistencies described above, clean
    the prefix.

    Examples:
        >>> inconsistent_text = "No. L-I9863"
        >>> gr_prefix_clean(inconsistent_text)
        'L-19863'

    Args:
        text (str): Raw docket serial ID that ought to be cleaned, e.g. ``L-I`
            or `No. L-`.

    Returns:
        str | None: The cleaned GR docket ID, if detected.
    """
    regex1, prefix1 = LEGACY_PREFIXED_LOOKALIKE, "L-1"  # L-I9863
    regex2, prefix2 = LEGACY_PREFIXED, "L-"  # improper L- formatted cases
    if cleaned1 := replace_prefix_regex(regex1, text, prefix1):
        return cleaned1
    elif cleaned2 := replace_prefix_regex(regex2, text, prefix2):
        return cleaned2
    return None



import abc
from collections.abc import Iterator
from typing import Self

from citation_report import Report
from pydantic import field_serializer

from .docket_category import DocketCategory
from .docket_model import Docket


class DocketReportCitation(Docket, Report, abc.ABC):
    """Note `Report` is defined in a separate library `citation-report`.

    The `DocketReportCitation` abstract class makes sure that all of the
    fields of a `Docket` object alongside all of the fields of a `Report`
    object will be utilized. It also mandates the implementation of a`cls.search()`
    method.
    """

    @property
    def _docket(self):
        if self.first_id:
            return f"{self.category.name} No. {self.serial_text}, {self.formatted_date}"
        elif self.ids:
            return f"{self.category.name} No. {self.ids}, {self.formatted_date}"
        return None

    @property
    def _report(self):
        return self.volpubpage or None

    def __str__(self) -> str:
        if self._docket and self._report:
            return f"{self._docket}, {self._report}"
        elif self._docket:
            return self._docket
        elif self._report:
            return self._report
        return "No citation."

    def __repr__(self) -> str:
        if self._docket and self._report:
            return f"<DocketReport: {self._docket} | {self._report}>"
        elif self._docket:
            return f"<DocketReport: {self._docket}>"
        elif self._report:
            return f"<DocketReport: {self._report}>"
        return "<DocketReport: improper citation>"

    @classmethod
    @abc.abstractmethod
    def search(cls, raw: str) -> Iterator[Self]:
        raise NotImplementedError("Search method must produce Iterator of instance.")



import re
from re import Pattern

formerly = r"""
    (?P<formerly>
        \.? # see A.M. No. MTJ-01-1381. (Formerly OCA I.P.I No. 97-426-MTJ)
        \s*
        [\[\(]
        \s*
        (FORMERLY|Formerly|Former|From)\:?
        (?P<old_name>[\w\s\-\.\&]+)
        \s*
        [\)\]]
        \W*
    )
"""

pp = r"""
    (?P<pp>
        \b
        (?P<extra_pages_label>
            p\.
            |
            pp\.
        )
        \s*
        \b
        (?P<extra_page_numbers_under_pp>
            [\d-]+
        )
        \b
        \W*
    )
"""


EXTRA: Pattern = re.compile(rf"{formerly}", re.I | re.X)


def cull_extra(text: str):
    """Remove text like `formerly... from` the given text string"""
    return EXTRA.sub("", text)



import re
from datetime import date
from typing import Self

from citation_date import DOCKET_DATE_FORMAT
from citation_report.main import is_eq
from pydantic import BaseModel, ConfigDict, Field, field_serializer

from .docket_category import DocketCategory
from .gr_clean import gr_prefix_clean

DB_SERIAL_NUM = re.compile(r"^(?P<serial>[a-z0-9-]+).*$")
"""Ideally, the docket id for the database should only be alpha-numeric, lowercase with dashes."""  # noqa: E501


class Docket(BaseModel):
    """
    The `Docket` is the modern identifier of a Supreme Court decision. This data structure
    however is not the final form of the identifier since that description belongs to the `Citation`
    and the `CountedCitation`.

    The purpose of this intermediate structure is that a `Docket` is often paired with a `Report`, which
    is the traditional identifier based on volume and page numbers. The pairing however is not
    mandatory, thus needed flexibility to create structures with the following combinations of
    the eventual Citation object:

    Citation | Docket | Report
    :--:|:--:|:--:
    has both docket and report | yes | yes
    only a docket | yes | no
    only a report | no | yes

    See docket_citation.DocketReportCitation to see structure of paired content.

    A `Docket` is based on a `category`, a `serial id`, and a `date`. Since the serial id
    may required

    Field | Type | Description
    --:|:--:|:--
    `context` | optional (str) | Full text matched by the regex pattern
    `category` | optional (DocketCategory) | Whether GR, AC, etc.
    `ids` | optional (str) | The serial number of the docket category
    `docket_date` | optional (date) | The date associated with the docket

    Sample Citation | Category | Serial | Date
    :-- |:--:|:--:|:--:
    _G.R. Nos. 138570, October 10, 2000_ | GR | 74910 | October 10, 2000
    _A.M. RTJ-12-2317 (Formerly OCA I.P.I. No. 10-3378-RTJ), Jan 1, 2000_ | AM | RTJ-12-2317 |Jan 1, 2000
    _A.C. No. 10179 (Formerly CBD 11-2985), March 04, 2014_ | AC | 10179 | Mar. 4, 2014
    """  # noqa: E501

    context: str = Field(..., description="Full text matched by regex pattern.")
    category: DocketCategory = Field(..., description="e.g. General Register, etc.")
    ids: str = Field(
        ..., description="This may be comma-separated, e.g. '12, 32, and 41'"
    )
    docket_date: date = Field(...)

    def __repr__(self) -> str:
        return f"<Docket: {self.category} {self.serial_text}, {self.formatted_date}>"

    def __str__(self) -> str:
        if self.serial_text:
            return (
                f"{self.category} No. {self.serial_text.upper()}, {self.formatted_date}"  # noqa: E501
            )
        return "No proper string detected."

    def __eq__(self, other: Self) -> bool:
        opt_1 = is_eq(self.category.name, other.category.name)
        opt_2 = is_eq(self.first_id, other.first_id)
        opt_3 = is_eq(self.docket_date.isoformat(), other.docket_date.isoformat())
        return all([opt_1, opt_2, opt_3])

    @property
    def slug(self):
        return "-".join(
            [self.category.name, self.serial_text, self.docket_date.isoformat()]
        )

    @property
    def serial_text(self) -> str:
        """From raw `ids`, get the `cleaned_ids`, and of these `cleaned_ids`,
            extract the `@first_id` found to deal with compound ids, e.g.
            ids separated by 'and' and ','

        Returns:
            str: Singular text identifier
        """
        if x := self.first_id or self.ids:
            x = x.rstrip("*•[]")
            if bits := x.split():
                if len(bits) > 1:
                    if bits[0].isalpha() and not bits[1].startswith("-"):
                        x = f"{bits[0]}-{bits[1]}"
                    elif bits[0].isalpha() and bits[1].startswith("-"):  # mtj -02-1466
                        x = f"{bits[0]}{bits[1]}"
                    elif bits[1].isalpha() and not bits[0].endswith("-"):
                        x = f"{bits[0]}-{bits[1]}"
                    elif bits[1].isalpha() and bits[0].endswith("-"):  # '14061- ret'
                        x = f"{bits[0]}{bits[1]}"
            if adjust := gr_prefix_clean(x):
                return adjust
        return x.split()[0]

    @property
    def first_id(self) -> str:
        """Get first bit from list of separated ids, when possible.

        Returns:
            str: First id found
        """

        def first_exists(char: str, text: str):
            """If a `char` exists in the `text`, split on this value."""
            return text.split(char)[0] if char in text else None

        for char in [" - ", "/", ",", ";", " and ", " AND ", "&"]:
            if res := first_exists(char, self.ids):
                return res
        return self.ids

    @property
    def formatted_date(self) -> str | None:
        if self.docket_date:
            return self.docket_date.strftime(DOCKET_DATE_FORMAT)
        return None

    @classmethod
    def check_serial_num(cls, text: str) -> bool:
        """If a serial number exists, ensure it meets criteria prior to row creation."""
        if DB_SERIAL_NUM.search(text.lower()):
            return True
        return False

    @classmethod
    def clean_serial(cls, text: str) -> str | None:
        """Criteria:

        1. Must be lowercased
        2. Characters that can be included `a-z`, `0-9`, `-`
        3. Must only contain a single alpha-numeric reference

        Args:
            text (str): Raw text to clean

        Returns:
            str: Cleaned serial text fit for database input.
        """
        text = text.lower()
        if " " in text:
            text = text.split()[0]
        if match := DB_SERIAL_NUM.search(text):
            if candidate := match.group("serial"):
                return candidate
        return None



from .constructor import CitationConstructor
from .docket_category import DocketCategory
from .docket_citation import DocketReportCitation
from .docket_model import DB_SERIAL_NUM, DOCKET_DATE_FORMAT, Docket
from .docket_rules import is_statutory_rule
from .gr_clean import gr_prefix_clean
from .misc.extra import cull_extra, formerly, pp
from .misc.num import Num



from enum import StrEnum


class DocketCategory(StrEnum):
    """Common docket references involving Philippine Supreme Court decisions.

    Name | Value
    :--|:--
    `GR` | General Register
    `AM` | Administrative Matter
    `AC` | Administrative Case
    `BM` | Bar Matter
    `PET` | Presidential Electoral Tribunal
    `OCA` | Office of the Court Administrator
    `JIB` | Judicial Integrity Board
    `UDK` | Undocketed

    ## Complications

    ### Legacy rules

    These categories do not always represent decisions. For instance,
    there are are `AM` and `BM` docket numbers that represent rules rather
    than decisions.

    ### Redocketed numbers

    From the Supreme Court Stylebook (p. 159, 2024):

    > 11.3.1. Redocketed numbers
    >
    > Some cases may have an undocketed (UDK) number and may be redocketed and assigned
    a General Register (G.R.) number upon payment of the required docket fees. Still
    other cases may have a docket number starting with OCA IPI or JIB and may be
    redocketed as Administrative Matters (A.M.), while Commission on Bar Discipline
    (CBD) cases may be redocketed as Administrative Cases (A.C.). These must still be
    reflected in all court resolutions, orders, and decisions. x x x
    """

    GR = "General Register"
    AM = "Administrative Matter"
    AC = "Administrative Case"
    BM = "Bar Matter"
    PET = "Presidential Electoral Tribunal"
    OCA = "Office of the Court Administrator"
    JIB = "Judicial Integrity Board"
    UDK = "Undocketed"

    def __str__(self):
        return self.name

    def __repr__(self) -> str:
        """Uses name of member `gr` instead of Enum default
        `<DocketCategory.GR: 'General Register'>`. It becomes to
        use the following conventions:

        Examples:
            >>> DocketCategory['GR']
            'GR'
            >>> DocketCategory.GR
            'GR'

        Returns:
            str: The value of the Enum name
        """
        return str.__repr__(self.name.upper())



import unicodedata
from collections.abc import Iterator
from dataclasses import dataclass

from citation_report import Report

from .dockets import (
    CitationAC,
    CitationAM,
    CitationBM,
    CitationGR,
    CitationJIB,
    CitationOCA,
    CitationPET,
    CitationUDK,
    DocketReport,
    is_statutory_rule,
)


@dataclass
class CitableDocument:
    """Creates three main reusable lists:

    list | concept
    :--:|:--:
    `@docketed_reports` | list of `DocketReportCitation` found in the text, excluding exceptional statutory dockets
    `@reports` | list of `Report` found in the text (which may already be included in `@docketed_reports`)
    `@undocketed_reports` | = `@docketed_reports` - `@reports`

    Examples:
        >>> text_statutes = "Bar Matter No. 803, Jan. 1, 2000; Bar Matter No. 411, Feb. 1, 2000"
        >>> len(CitableDocument(text=text_statutes).docketed_reports) # no citations, since these are 'statutory dockets'
        0
        >>> text_cites = "374 Phil. 1, 10-11 (1999) 1111 SCRA 1111; G.R. No. 147033, April 30, 2003; G.R. No. 147033, April 30, 2003, 374 Phil. 1, 600; ABC v. XYZ, G.R. Nos. 138570, 138572, 138587, 138680, 138698, October 10, 2000, 342 SCRA 449;  XXX, G.R. No. 31711, Sept. 30, 1971, 35 SCRA 190; Hello World, 1111 SCRA 1111; Y v. Z, 35 SCRA 190;"
        >>> doc1 = CitableDocument(text=text_cites)
        >>> len(doc1.docketed_reports)
        4
        >>> doc1.undocketed_reports
        {'1111 SCRA 1111'}
        >>> text = "<em>Gatchalian Promotions Talent Pool, Inc. v. Atty. Naldoza</em>, 374 Phil. 1, 10-11 (1999), citing: <em>In re Almacen</em>, 31 SCRA 562, 600 (1970).; People v. Umayam, G.R. No. 147033, April 30, 2003; <i>Bagong Alyansang Makabayan v. Zamora,</i> G.R. Nos. 138570, 138572, 138587, 138680, 138698, October 10, 2000, 342 SCRA 449; Villegas <em>v.</em> Subido, G.R. No. 31711, Sept. 30, 1971, 41 SCRA 190;"
        >>> doc2 = CitableDocument(text=text)
        >>> set(doc2.get_citations()) == {'GR No. 147033, Apr. 30, 2003', 'GR No. 138570, Oct. 10, 2000, 342 SCRA 449', 'GR No. 31711, Sep. 30, 1971, 41 SCRA 190', '374 Phil. 1', '31 SCRA 562'}
        True
    """  # noqa: E501

    text: str

    def __post_init__(self):
        self.text = unicodedata.normalize("NFKD", self.text)
        self.reports = list(Report.extract_reports(self.text))
        self.docketed_reports = list(self.get_docketed_reports(self.text))
        self.undocketed_reports = self.get_undocketed_reports()

    @classmethod
    def get_docketed_reports(
        cls, text: str, exclude_docket_rules: bool = True
    ) -> Iterator[DocketReport]:
        """Extract from `raw` text all raw citations which should include their `Docket` and `Report` component parts.
        This may however include statutory rules since some docket categories like AM and BM use this convention.
        To exclude statutory rules, a flag is included as a default.

        Examples:
            >>> cite = next(CitableDocument.get_docketed_reports("Bagong Alyansang Makabayan v. Zamora, G.R. Nos. 138570, 138572, 138587, 138680, 138698, October 10, 2000, 342 SCRA 449"))
            >>> cite.model_dump(exclude_none=True)
            {'publisher': 'SCRA', 'volume': '342', 'page': '449', 'context': 'G.R. Nos. 138570, 138572, 138587, 138680, 138698', 'category': 'GR', 'ids': '138570, 138572, 138587, 138680, 138698', 'docket_date': datetime.date(2000, 10, 10)}
            >>> statutory_text = "Bar Matter No. 803, Jan. 1, 2000"
            >>> next(CitableDocument.get_docketed_reports(statutory_text)) # default
            Traceback (most recent call last):
                ...
            StopIteration

        Args:
            text (str): Text to look for `Dockets` and `Reports`

        Yields:
            Iterator[DocketReport]: Any of custom `Docket` with `Report` types, e.g. `CitationAC`, etc.
        """  # noqa: E501
        text = unicodedata.normalize("NFKD", text)
        for search_func in (
            CitationAC.search,
            CitationAM.search,
            CitationOCA.search,
            CitationBM.search,
            CitationGR.search,
            CitationPET.search,
            CitationUDK.search,
            CitationJIB.search,
        ):
            # Each search function is applied to the text, each match yielded
            for result in search_func(text):
                if exclude_docket_rules:
                    if is_statutory_rule(result):
                        continue
                    yield result

    def get_undocketed_reports(self):
        """Steps:

        1. From a set of `uniq_reports` (see `self.reports`);
        2. Compare to reports found in `@docketed_reports`
        3. Limit reports to those _without_ an accompaying docket
        """
        uniq_reports = set(Report.get_unique(self.text))
        for cite in self.docketed_reports:
            if cite.volpubpage in uniq_reports:
                uniq_reports.remove(cite.volpubpage)
        return uniq_reports

    def get_citations(self) -> Iterator[str]:
        """There are two main lists to evaluate:

        1. `@docketed_reports` - each includes a `Docket` (optionally attached to a `Report`)
        2. `@reports` - from the same text, just get `Report` objects.

        Can filter out `Report` objects not docketed and thus return
        a more succinct citation list which includes both constructs mentioned above but
        without duplicate `reports`.
        """  # noqa: E501
        if self.docketed_reports:
            for doc_report_cite in self.docketed_reports:
                yield str(doc_report_cite)

            if self.undocketed_reports:
                yield from self.undocketed_reports  # already <str>
        else:
            if self.reports:
                for report in self.reports:
                    yield str(report)



import re
from enum import Enum

from .docket_category import DocketCategory
from .docket_model import Docket


class DocketRuleSerialNumber(Enum):
    """A rule is more adequately defined in `statute-utils`. There are are `AM` and `BM` docket numbers
    that represent rules rather than decisions. The statutes are contained in a folder, e.g.
    `/corpus-statutes/rule-am`, `/corpus-statutes/rule-bm` and thus may be extracted likeso:

    ```py
    from pathlib import Path
    statutes = Path().home() / "code" / "corpus-statutes"
    ams = statutes.glob("rule_am/*")
    [am.stem for am in ams]
    ```

    """  # noqa: E501

    BarMatter = [
        803,
        1922,
        1645,
        850,
        287,
        1132,
        1755,
        1960,
        209,
        1153,
        411,
        356,
    ]
    AdminMatter = [
        "00-2-10-SC",
        "10-4-20-SC",
        "02-9-02-SC",
        "19-08-15-SC",
        "07-7-12-SC",
        "02-8-13-SC",
        "02-11-10-SC",
        "04-10-11-SC",
        "03-06-13-SC",
        "19-10-20-SC",
        "99-2-02-SC",
        "02-11-11-SC",
        "12-12-11-SC",
        "01-7-01-SC",
        "00-5-03-SC",
        "07-4-15-SC",
        "02-2-07-SC",
        "01-1-03-SC",
        "02-11-12-SC",
        "19-03-24-SC",
        "02-6-02-SC",
        "03-04-04-SC",
        "03-1-09-SC",
        "08-1-16-SC",
        "15-08-02-SC",
        "99-10-05-0",
        "06-11-5-SC",
        "03-02-05-SC",
        "00-4-07-SC",
        "00-8-10-SC",
        "04-2-04-SC",
        "12-8-8-SC",
        "21-08-09-SC",
        "03-05-01-SC",
        "09-6-8-SC",
        "05-8-26-SC",
        "00-2-03-SC",
        "01-8-10-SC",
    ]

    @property
    def regex(self) -> str:
        return r"(?:" + "|".join(str(i) for i in self.value) + r")"

    @property
    def pattern(self) -> re.Pattern:
        return re.compile(self.regex, re.I)


StatutoryBM = DocketRuleSerialNumber.BarMatter.pattern
"""Fixed regex compiled pattern for Statutory Bar Matter"""

StatutoryAM = DocketRuleSerialNumber.AdminMatter.pattern
"""Fixed regex compiled pattern for Statutory Administrative Matter"""


def is_statutory_rule(citeable):
    """Determine if `citeable` object is a statutory pattern based on a specific
    lising of `category` and `serial_text`.

    Examples:
        >>> from dateutil.parser import parse
        >>> test_num = '19-03-24-SC' # a statute, see list
        >>> d0 = Docket(context="", category=DocketCategory['AM'], ids=test_num, docket_date=parse("Jan. 1, 2022"))
        >>> is_statutory_rule(d0)
        True
        >>> test_num = '19-03-24-sc' # a statute, see list
        >>> d1 = Docket(context="", category=DocketCategory['AM'], ids=test_num, docket_date=parse("Jan. 1, 2022"))
        >>> is_statutory_rule(d1)
        True
        >>> test_num = '19-03-00-sc' # NOT A STATUTE
        >>> d2 = Docket(context="", category=DocketCategory['AM'], ids=test_num, docket_date=parse("Jan. 1, 2022"))
        >>> is_statutory_rule(d2)
        False
    """  # noqa: E501

    if isinstance(citeable, Docket):  # excludes solo reports
        if citeable.category == DocketCategory.BM:
            if StatutoryBM.search(citeable.serial_text):
                return True
        elif citeable.category == DocketCategory.AM:
            if StatutoryAM.search(citeable.serial_text):
                return True
    return False



import datetime
import logging
from collections.abc import Iterator
from functools import cached_property
from typing import Self

from citation_date import DOCKET_DATE_FORMAT
from citation_report import Report
from dateutil.parser import parse
from pydantic import BaseModel, ConfigDict, Field, field_serializer, model_serializer

from .dockets import Docket, DocketCategory
from .document import CitableDocument


class Citation(BaseModel):
    """
    A Philippine Supreme Court `Citation` consists of:

    1. `Docket` includes:
        1. _category_,
        2. _serial number_, and
        3. _date_.
    2. `Report` - as defined in [citation-report](https://github.com/justmars/citation-report) - includes:
        1. _volume number_,
        2. _identifying acronym of the reporter/publisher_,
        3. _page of the reported volume_.

    It is typical to see a `Docket` combined with a `Report`:

    > _Bagong Alyansang Makabayan v. Zamora, G.R. Nos. 138570, 138572, 138587, 138680, 138698, October 10, 2000, 342 SCRA 449_

    Taken together (and using _Bagong Alyansang Makabayan_ as an example) the text above can be extracted into fields:

    Example | Field | Type | Description
    --:|:--:|:--|--:
    GR | `docket_category` | optional (`ShortDocketCategory`) | See shorthand
    138570 |`docket_serial` | optional (str) | See serialized identifier
    datetime.date(2000, 10, 10) | `docket_date` | optional (date) | When docket serial issued
    GR 138570, Oct. 10, 2000 | `docket` | optional (str) | Combined `docket_category` `docket_serial` `docket_date`
    None | `phil` | optional (str) | combined `volume` Phil. `page`
    342 SCRA 449 | `scra` | optional (str) | combined `volume` SCRA `page`
    None | `offg` | optional (str) | combined `volume` O.G. `page`
    """  # noqa: E501

    model_config = ConfigDict(str_strip_whitespace=True)
    docket_category: DocketCategory | None = Field(default=None, alias="cat")
    docket_serial: str | None = Field(default=None, alias="num")
    docket_date: datetime.date | None = Field(default=None, alias="date")
    phil: str | None = Field(default=None)
    scra: str | None = Field(default=None)
    offg: str | None = Field(default=None)

    def __repr__(self) -> str:
        return f"<Citation: {str(self)}>"

    def __str__(self) -> str:
        docket_str = self.get_docket_display()
        report_str = self.phil or self.scra or self.offg
        if docket_str and report_str:
            return f"{docket_str}, {report_str}"
        elif docket_str:
            return f"{docket_str}"
        elif report_str:
            return f"{report_str}"
        return f"<Bad citation str: {self.docket_category=} {self.docket_serial=} {self.docket_date=}>"  # noqa: E501

    def __eq__(self, other: Self) -> bool:
        """Helps `seen` variable in `CountedCitation`: either the docket bits match
        or any of the report fields match."""

        def is_docket_match(other: Self) -> bool:
            """All the docket elements must match to be equal."""
            cat_is_eq = (
                self.docket_category is not None
                and other.docket_category is not None
                and (self.docket_category == other.docket_category)
            )
            num_is_eq = (
                self.docket_serial is not None
                and other.docket_serial is not None
                and (self.docket_serial == other.docket_serial)
            )
            date_is_eq = (
                self.docket_date is not None
                and other.docket_date is not None
                and (self.docket_date == other.docket_date)
            )
            return all([cat_is_eq, num_is_eq, date_is_eq])

        if is_docket_match(other):
            return True

        return any(
            [
                self.scra and other.scra and self.scra.lower() == other.scra.lower(),
                self.phil and other.phil and self.phil.lower() == other.phil.lower(),
                self.offg and other.offg and self.offg.lower() == other.offg.lower(),
            ]
        )

    @field_serializer("docket_date")
    def serialize_dt(self, value: datetime.date | None):
        if value:
            return value.isoformat()

    @field_serializer("docket_serial")
    def serialize_num(self, value: str | None):
        if value:
            return Docket.clean_serial(value)

    @field_serializer("docket_category")
    def serialize_cat(self, value: DocketCategory | None):
        if value:
            return value.name.lower()

    @field_serializer("phil")
    def serialize_phil(self, value: str | None):
        if value:
            return value.lower()

    @field_serializer("scra")
    def serialize_scra(self, value: str | None):
        if value:
            return value.lower()

    @field_serializer("offg")
    def serialize_offg(self, value: str | None):
        if value:
            return value.lower()

    @model_serializer
    def ser_model(self) -> dict[str, str | datetime.date | None]:
        """Generate a database row-friendly format of the model. Note the different
        field names: `cat`, `num`, `dt`, `phil`, `scra`, `offg` map to either a usable
        database value or `None`. The docket values here have the option to be `None`
        since some citations, especially the legacy variants, do not include their
        docket equivalents in the source texts.

        Examples:
            >>> text = "OCA IPI No. 10-3450-P, Feb. 06, 2008"
            >>> cite = Citation.extract_citation(text)
            >>> cite.model_dump_json()
            '{"cat":"oca","num":"10-3450-p","date":"2008-02-06","phil":null,"scra":null,"offg":null}'

        """
        return {
            "cat": self.serialize_cat(self.docket_category),
            "num": self.serialize_num(self.docket_serial),
            "date": self.serialize_dt(self.docket_date),
            "phil": self.serialize_phil(self.phil),
            "scra": self.serialize_scra(self.scra),
            "offg": self.serialize_offg(self.offg),
        }

    def set_slug(self) -> str | None:
        """Create a unique identifier of a decision.

        Examples:
            >>> text = "GR 138570, Oct. 10, 2000"
            >>> Citation.extract_citation(text).set_slug()
            'gr-138570-2000-10-10'

        """
        bits = [
            self.serialize_cat(self.docket_category),
            self.serialize_num(self.docket_serial),
            self.serialize_dt(self.docket_date),
        ]
        if all(bits):
            return "-".join(bits)  # type: ignore
        return None

    @classmethod
    def get_docket_slug_from_text(cls, v: str) -> str | None:
        """Given a docket string, format the string into a slug
        that has the same signature as a database primary key.

        Examples:
            >>> text = "GR 138570, Oct. 10, 2000"
            >>> Citation.get_docket_slug_from_text(text)
            'gr-138570-2000-10-10'

        Args:
            v (str): The text to evaluate

        Returns:
            str | None: The slug to use, if possible.
        """
        if cite := cls.extract_citation(v):
            if cite.is_docket:
                return cite.set_slug()
        return None

    def make_docket_row(self):
        """This presumes that a valid docket exists. Although a citation can
        be a non-docket, e.g. phil, scra, etc., for purposes of creating a
        a route-based row for a prospective decision object, the identifier will be
        based on a docket id."""
        if id := self.set_slug():
            return self.model_dump() | {"id": id}
        logging.error(f"Undocketable: {self=}")
        return None

    @classmethod
    def from_docket_row(
        cls,
        cat: str,
        num: str,
        date: str,
        opt_phil: str | None,
        opt_scra: str | None,
        opt_offg: str | None,
    ):
        return cls(
            cat=DocketCategory[cat.upper()],
            num=num,
            date=parse(date).date(),
            phil=cls.get_report(opt_phil),
            scra=cls.get_report(opt_scra),
            offg=cls.get_report(opt_offg),
        )

    @cached_property
    def is_docket(self) -> bool:
        return all([self.docket_category, self.docket_serial, self.docket_date])

    @cached_property
    def display_date(self):
        """This is the same as Docket@formatted_date."""
        if self.docket_date:
            return self.docket_date.strftime(DOCKET_DATE_FORMAT)
        return None

    def get_docket_display(self) -> str | None:
        if self.is_docket:
            return (
                f"{self.docket_category} No. {self.docket_serial}, {self.display_date}"  # type: ignore # noqa: E501
            )
        return None

    @classmethod
    def get_report(cls, raw: str | None = None) -> str | None:
        """Get a lower cased volpubpage of `publisher` from the `data`. Assumes
        that the publisher key is either `phil`, `scra` or `offg`.

        Examples:
            >>> raw = "123 Phil. 123"
            >>> Citation.get_report(raw)
            '123 phil. 123'

        Args:
            raw (str): Text to examine

        Returns:
            str | None: _description_
        """
        if not raw:
            return None

        try:
            reports = Report.extract_reports(raw)
            report = next(reports)
            if result := report.volpubpage:
                return result.lower()
            else:
                logging.warning(f"No volpubpage {raw=}")
                return None
        except StopIteration:
            logging.warning(f"No {raw=} report")
            return None

    @classmethod
    def _set_report(cls, text: str):
        try:
            obj = next(Report.extract_reports(text))
            return cls(phil=obj.phil, scra=obj.scra, offg=obj.offg)
        except StopIteration:
            logging.debug(f"{text} is not a Report instance.")
            return None

    @classmethod
    def _set_docket_report(cls, text: str):
        try:
            obj = next(CitableDocument.get_docketed_reports(text))
            return cls(
                cat=obj.category,
                num=obj.serial_text,
                date=obj.docket_date,
                phil=obj.phil,
                scra=obj.scra,
                offg=obj.offg,
            )
        except StopIteration:
            logging.debug(f"{text} is not a Docket nor a Report instance.")
            return None

    @classmethod
    def extract_citations(cls, text: str) -> Iterator[Self]:
        """Find citations and parse resulting strings to determine whether they are:

        1. `Docket` + `Report` objects (in which case, `_set_docket_report()` will be used); or
        2. `Report` objects (in which case `_set_report()` will be used)

        Then processing each object so that they can be structured in a uniform format.

        Examples:
            >>> text = "<em>Gatchalian Promotions Talent Pool, Inc. v. Atty. Naldoza</em>, 374 Phil. 1, 10-11 (1999), citing: <em>In re Almacen</em>, 31 SCRA 562, 600 (1970).; People v. Umayam, G.R. No. 147033, April 30, 2003; <i>Bagong Alyansang Makabayan v. Zamora,</i> G.R. Nos. 138570, 138572, 138587, 138680, 138698, October 10, 2000, 342 SCRA 449; Villegas <em>v.</em> Subido, G.R. No. 31711, Sept. 30, 1971, 41 SCRA 190;"
            >>> len(list(Citation.extract_citations(text)))
            5

        Args:
            text (str): Text to evaluate

        Yields:
            Iterator[Self]: Itemized citations pre-processed via `CitableDocument`
        """  # noqa: E501
        for cite in CitableDocument(text=text).get_citations():
            if _docket := cls._set_docket_report(cite):
                yield _docket
            elif _report := cls._set_report(cite):
                yield _report
            else:
                logging.error(f"Skip invalid {cite=}.")

    @classmethod
    def extract_citation(cls, text: str) -> Self | None:
        """Thin wrapper over `cls.extract_citations()`.

        Examples:
            >>> Citation.extract_citation('Hello World') is None
            True
            >>> next(Citation.extract_citations('12 Phil. 24'))
            <Citation: 12 Phil. 24>

        Args:
            text (str): Text to evaluate

        Returns:
            Self | None: First item found from `extract_citations`, if it exists.
        """
        try:
            return next(cls.extract_citations(text))
        except StopIteration:
            return None

    @classmethod
    def make_citation_string(
        cls,
        cat: str,
        num: str,
        date: str,
        phil: str | None = None,
        scra: str | None = None,
        offg: str | None = None,
    ) -> str | None:
        """Assume that because of citation-utils, the extracted values are inputted into a database.

        When the values are pulled from the database, it becomes necessary to convert these database (lowercased) values
        to a unified properly-cased citation string with readable date (vs. isoformat db-counterpart).

        Examples:
            >>> Citation.make_citation_string(cat='gr', num='111', date='2000-01-01', phil='100 phil. 100', scra='122 scra 100-a')
            'G.R. No. 111, Jan. 1, 2000, 100 Phil. 100, 122 SCRA 100-A'

        Args:
            cat (str): The shorthand code for docket category
            num (str): The serial identifier of the docket category
            date (str): The date of `cat` + `num`
            phil (str | None, optional): Phil. Reports. Defaults to None.
            scra (str | None, optional): Supreme Court Reports Annotated. Defaults to None.
            offg (str | None, optional): Official Gazette. Defaults to None.

        Returns:
            str | None: The combination of citation bits.
        """  # noqa: E501
        bits = [
            phil.title().split() if phil else None,
            scra.upper().split() if scra else None,
            offg.upper().split() if offg else None,
        ]
        cased_bits = [
            f"{bit[0].upper()} {bit[1]} {bit[2].upper()}" for bit in bits if bit
        ]
        reports = ", ".join(cased_bits) if any(bits) else None
        dt = datetime.date.fromisoformat(date).strftime("%b. %-d, %Y")
        match cat:
            case "gr":
                prefix = "G.R."
            case "am":
                prefix = "A.M."
            case "ac":
                prefix = "A.C."
            case "bm":
                prefix = "B.M."
            case "udk":
                prefix = "UDK"
            case "jib":
                prefix = "JIB-FPI"
            case "oca":
                prefix = "OCA IPI"
            case "pet":
                prefix = "P.E.T."
            case _:
                return None
        docket = f"{prefix} No. {num.upper()}"
        bits = [bit for bit in [docket, dt, reports] if bit]
        return ", ".join(bits)


class CountedCitation(Citation):
    mentions: int = Field(default=1, description="Get count via Citation __eq__")

    def __repr__(self) -> str:
        return f"{str(self)}: {self.mentions}"

    @model_serializer
    def ser_model(self) -> dict[str, str | datetime.date | int | None]:
        return {
            "cat": self.serialize_cat(self.docket_category),
            "num": self.serialize_num(self.docket_serial),
            "date": self.serialize_dt(self.docket_date),
            "phil": self.serialize_phil(self.phil),
            "scra": self.serialize_scra(self.scra),
            "offg": self.serialize_offg(self.offg),
            "mentions": self.mentions,
        }

    @classmethod
    def from_source(cls, text: str) -> list[Self]:
        """Computes mentions of `counted_dockets()` vis-a-vis `counted_reports()` and
        count the number of unique items, taking into account the Citation
        structure and the use of __eq__ re: what is considered unique.

        Examples:
            >>> source = "374 Phil. 1, 10-11 (1999) 1111 SCRA 1111; G.R. No. 147033, April 30, 2003; G.R. No. 147033, April 30, 2003, 374 Phil. 1, 600; ABC v. XYZ, G.R. Nos. 138570, 138572, 138587, 138680, 138698, October 10, 2000, 342 SCRA 449;  XXX, G.R. No. 31711, Sept. 30, 1971, 35 SCRA 190; Hello World, 1111 SCRA 1111; Y v. Z, 35 SCRA 190; 1 Off. Gaz. 41 Bar Matter No. 803, Jan. 1, 2000 Bar Matter No. 411, Feb. 1, 2000 Bar Matter No. 412, Jan. 1, 2000, 1111 SCRA 1111; 374 Phil. 1"
            >>> len(CountedCitation.from_source(source))
            5

        Args:
            text (str): Text to Evaluate.

        Returns:
            list[Self]: Unique citations with their counts.
        """  # noqa: E501
        all_reports = cls.counted_reports(text)  # includes reports in docket_reports
        if drs := cls.counted_docket_reports(text):
            for dr in drs:
                for report in all_reports:
                    if report == dr:  # uses Citation __eq__
                        balance = 0
                        if report.mentions > dr.mentions:
                            balance = report.mentions - dr.mentions
                        dr.mentions = dr.mentions + balance
                        report.mentions = 0
            return drs + [report for report in all_reports if report.mentions > 0]
        return all_reports

    @classmethod
    def from_repr_format(cls, repr_texts: list[str]) -> Iterator[Self]:
        """Generate their pydantic counterparts from `<cat> <id>: <mentions>` format.

        Examples:
            >>> repr_texts = ['BM No. 412, Jan 01, 2000, 1111 SCRA 1111: 3', 'GR No. 147033, Apr 30, 2003, 374 Phil. 1: 1']
            >>> results = list(CountedCitation.from_repr_format(repr_texts))
            >>> len(results)
            2
            >>> results[0].model_dump()
            {'cat': 'bm', 'num': '412', 'date': '2000-01-01', 'phil': None, 'scra': '1111 scra 1111', 'offg': None, 'mentions': 3}
            >>> results[1].model_dump()
            {'cat': 'gr', 'num': '147033', 'date': '2003-04-30', 'phil': '374 phil. 1', 'scra': None, 'offg': None, 'mentions': 1}

        Args:
            repr_texts (str): list of texts having `__repr__` format of a `CountedRule`

        Yields:
            Iterator[Self]: Instances of CountedCitation.
        """  # noqa: E501
        for text in repr_texts:
            counted_bits = text.split(":")
            if len(counted_bits) == 2:
                if cite := cls.extract_citation(counted_bits[0].strip()):
                    citation = cls(
                        cat=cite.docket_category,
                        num=cite.docket_serial,
                        date=cite.docket_date,
                        phil=cite.phil,
                        scra=cite.scra,
                        offg=cite.offg,
                    )
                    citation.mentions = int(counted_bits[1].strip())
                    yield citation

    @classmethod
    def counted_reports(cls, text: str):
        """Detect _reports_ only from source `text` by first converting
        raw citations into a `Citation` object to take advantage of `__eq__` in
        a `seen` list. This will also populate the the unique records with missing
        values.
        """
        seen: list[cls] = []  # type: ignore
        reports = Report.extract_reports(text=text)
        for report in reports:
            cite = Citation(phil=report.phil, scra=report.scra, offg=report.offg)
            if cite not in seen:
                seen.append(cls(**cite.model_dump()))
            else:
                included = seen[seen.index(cite)]
                included.mentions += 1
        return seen

    @classmethod
    def counted_docket_reports(cls, text: str):
        """Detect _dockets with reports_ from source `text` by first converting
        raw citations into a `Citation` object to take advantage of `__eq__` in
        a `seen` list. Will populate unique records with missing values.
        """

        seen: list[cls] = []  # type: ignore
        for obj in CitableDocument.get_docketed_reports(text=text):
            cite = Citation(
                cat=obj.category,
                num=obj.serial_text,
                date=obj.docket_date,
                phil=obj.phil,
                scra=obj.scra,
                offg=obj.offg,
            )
            if cite not in seen:
                seen_citation = cls(
                    cat=cite.docket_category,
                    num=cite.docket_serial,
                    date=cite.docket_date,
                    phil=cite.phil,
                    scra=cite.scra,
                    offg=cite.offg,
                )
                seen.append(seen_citation)
            else:
                included = seen[seen.index(cite)]
                included.mentions += 1
                included.add_values(cite)  # for citations, can add missing
        return seen

    def add_values(self, other: Citation):
        if not self.docket_category and other.docket_category:
            self.docket_category = other.docket_category

        if not self.docket_serial and other.docket_serial:
            self.docket_serial = other.docket_serial

        if not self.docket_date and other.docket_date:
            self.docket_date = other.docket_date

        if not self.scra and other.scra:
            self.scra = other.scra

        if not self.phil and other.phil:
            self.phil = other.phil

        if not self.offg and other.offg:
            self.offg = other.offg



from collections.abc import Iterator
from typing import Self

from .models import CitationConstructor, DocketCategory, DocketReportCitation
from .models.misc import NUMBER_KEYWORD

pet_key = rf"""
    (
        P
        [\.\s]*
        E
        [\.\s]*
        T
        [\.\s]*
        (Case)?
    )
    \s+
    {NUMBER_KEYWORD}
"""


required = rf"""
    (?P<pet_init>
        {pet_key}
    )
    (?P<pet_middle>
        \d+
    )
"""


pet_phrases = rf"""
    (?P<pet_phrase>
        {required}
        [\,\s]*
    )
"""

constructed_pet = CitationConstructor(
    label=DocketCategory.PET.value,
    short_category=DocketCategory.PET.name,
    group_name="pet_phrase",
    init_name="pet_init",
    docket_regex=pet_phrases,
    key_regex=pet_key,
    num_regex=NUMBER_KEYWORD,
)


class CitationPET(DocketReportCitation):
    ...

    @classmethod
    def search(cls, text: str) -> Iterator[Self]:
        """Get all dockets matching the `PET` docket pattern, inclusive of their optional Report object.

        Examples:
            >>> text = "P.E.T. Case No. 001, February 13, 1996"
            >>> cite = next(CitationPET.search(text))
            >>> cite.model_dump(exclude_none=True)
            {'context': 'P.E.T. Case No. 001', 'category': 'PET', 'ids': '001', 'docket_date': datetime.date(1996, 2, 13)}

        Args:
            text (str): Text to look for citation objects

        Yields:
            Iterator[Self]: Combination of Docket and Report pydantic model.
        """  # noqa E501
        for result in constructed_pet.detect(text):
            yield cls(**result)



from .citation import Citation, CountedCitation
from .dockets import (
    DOCKET_DATE_FORMAT,
    CitationAC,
    CitationAM,
    CitationBM,
    CitationGR,
    CitationJIB,
    CitationOCA,
    CitationPET,
    CitationUDK,
    Docket,
    DocketCategory,
    Num,
    ac_key,
    ac_phrases,
    am_key,
    am_phrases,
    bm_key,
    bm_phrases,
    cull_extra,
    formerly,
    gr_key,
    gr_phrases,
    is_statutory_rule,
    jib_key,
    jib_phrases,
    l_key,
    oca_key,
    oca_phrases,
    pet_key,
    pet_phrases,
    pp,
    udk_key,
    udk_phrases,
)
from .document import CitableDocument
from .special import extract_docket_meta



import re
from typing import Any

from dateutil.parser import parse

from .document import CitableDocument

## Used for a specific website page

DOCKET_PATTERN = re.compile(
    r"""
    (?P<docket>.*?)
    \.
    \s+
    (?P<decision_date>[A-Z]\w+\s\d+,\s\d{4})
    (\s+
    \[
        Date
        \s
        Uploaa?ded: # deal with mispelled word
        \s
        (?P<upload_date>.*?)
    \])? # optional upload date, see Opinions which do not have them
""",
    re.X,
)


def extract_docket_meta(text: str) -> dict[str, Any] | None:
    """Metadata involving the docket.

    Examples:
        >>> extract_docket_meta("G.R. No. 234179. December 5, 2022 [Date Uploaded: 01/26/2023]")
        {'docket': 'G.R. No. 234179', 'decision_date': datetime.date(2022, 12, 5), 'upload_date': '01/26/2023', 'context': 'G.R. No. 234179', 'category': 'GR', 'ids': '234179', 'docket_date': datetime.date(2022, 12, 5)}
        >>> extract_docket_meta("G.R. No. 180350/G.R. No. 205186/G.R. No. 222919/G.R. No. 223237. July 6, 2022 [Date Uploaded: 09/21/2022]")
        {'docket': 'G.R. No. 180350/G.R. No. 205186/G.R. No. 222919/G.R. No. 223237', 'decision_date': datetime.date(2022, 7, 6), 'upload_date': '09/21/2022', 'context': 'G.R. No. 180350/G.R. No. 205186/G.R. No. 222919/G.R. No. 223237', 'category': 'GR', 'ids': '180350', 'docket_date': datetime.date(2022, 7, 6)}

    Args:
        text (str): Line from Supreme Court URL sub-title heading.

    Returns:
        dict[str, Any] | None: Docket-based details.
    """  # noqa: E501
    match = DOCKET_PATTERN.search(text)
    if not match:
        return None
    res = match.groupdict()

    if not res.get("docket"):
        return None

    if not (_date := res.get("decision_date")):
        return None

    try:
        res["decision_date"] = parse(_date, fuzzy=True).date()
        format_date = res["decision_date"].strftime("%B %d, %Y")
        citation_text = f"{res['docket']}, {format_date}"
        citation_obj = next(CitableDocument.get_docketed_reports(citation_text))
        cleaned_id = citation_obj.first_id
        res |= citation_obj.model_dump(
            exclude={
                "volume",
                "publisher",
                "page",
                "report",
                "report_date",
                "volpubpage",
            }
        )
        if res.get("ids") is None:
            return None
        else:
            res["ids"] = cleaned_id
    except Exception:
        res["decision_date"] = None
        return None
    return res



from collections.abc import Iterator
from typing import Self

from .models import CitationConstructor, DocketCategory, DocketReportCitation, Num

separator = r"[,\.\s-]*"
l_digits = r"\bL\-\d+"
digits_alone = r"\d+"

acronyms = r"""
    \s?
    (
        SBC # B.M. SBC-591, December 1, 1977
    )
    \s?
"""
letter = rf"""
    (
        \b
        {acronyms}
    )?
    [\d-]{{3,}} #  at least two digits and a dash
    ( # don't add \b  to capture "-Ret.""
        {acronyms}
    )?
"""

bm_key = rf"""
    (
        (
            b
            {separator}
            m
            {separator}
        )|
        (
            \b
            bar
            \s+
            matter
            \s*
        )
    )
"""

bm_num = rf"""
    (
        {bm_key}
        {Num.BM.allowed}
    )
"""

required = rf"""
    (?P<bm_init>
        {bm_num}
    )
    (?P<bm_middle>
        (
            ({letter})|
            ({l_digits})|
            ({digits_alone})
        )
    )
    (?:
        (
            [\,\s,\-\&]|
            and
        )*
    )?
"""

optional = rf"""
    (?P<bm_init_optional>
        {bm_num}
    )?
    (?P<bm_middle_optional>
        ({letter})|
        ({l_digits})|
        ({digits_alone})
    )?
    (?:
        (
            [\,\s,\-\&]|
            and
        )*
    )?
"""

bm_phrases = rf"""
    (?P<bm_phrase>
        ({required})
        ({optional}){{1,3}}
    )
"""


constructed_bm = CitationConstructor(
    label=DocketCategory.BM.value,
    short_category=DocketCategory.BM.name,
    group_name="bm_phrase",
    init_name="bm_init",
    docket_regex=bm_phrases,
    key_regex=bm_key,
    num_regex=Num.BM.allowed,
)


class CitationBM(DocketReportCitation):
    ...

    @classmethod
    def search(cls, text: str) -> Iterator[Self]:
        """Get all dockets matching the `BM` docket pattern, inclusive of their optional Report object.

        Examples:
            >>> text = "B.M. No. 1678, December 17, 2007"
            >>> cite = next(CitationBM.search(text))
            >>> cite.model_dump(exclude_none=True)
            {'context': 'B.M. No. 1678', 'category': 'BM', 'ids': '1678', 'docket_date': datetime.date(2007, 12, 17)}

        Args:
            text (str): Text to look for citation objects

        Yields:
            Iterator[Self]: Combination of Docket and Report pydantic model.
        """  # noqa E501
        for result in constructed_bm.detect(text):
            yield cls(**result)



from collections.abc import Iterator
from typing import Self

from .models import CitationConstructor, DocketCategory, DocketReportCitation, Num

separator = r"[,\.\s-]*"
digit = r"\d{3}[\d-]*"  # e.g. 323-23, 343-34
acronyms = r"""
    \s?
    (
        A|
        B|
        P|
        R|
        P|
        J|
        OCA|
        O\.C\.A\.|
        SDC|
        CTA|
        RTC|
        CFI|
        RTJ|
        MTJ|
        MJ|
        CJ|
        CTJ|
        CAR|
        CCC|
        CFI|
        JDRC|
        OMB|
        TEL|
        METC|
        MCTC|
        MTCC|
        MTC|
        HOJ|
        RET(?:\.)?|
        SCC?|
        SC(?:\-PHILJA)?|
        CA(?:\-J)?|
        SB(?:\-J)?
    )
    \s?
"""
letter = rf"""
    (
        \b
        {acronyms}
    )?
    [\d-]{{3,}} #  at least two digits and a dash
    ( # don't add \b  to capture "-Ret.""
        {acronyms}
    )?
"""


ipi = r"""
    (
        (IPI)|(I\.P\.I\.)
    )
"""

am_key = rf"""
    (
        (
            a
            {separator}
            m
            {separator}
        )|
        (
            \b
            adm(in)?(istrative)?
            {separator}
            (?:
                \b
                (
                    Matter|
                    Mat\. # ADM. MAT. NO. P-97-1241
                )
                \s* # optional space
            )?
        )
    )
"""


am_num = rf"""
    (
        {am_key}
        {Num.AM.allowed}
    )
"""

required = rf"""
    (?P<am_init>
        {am_num}| # AM No.
        {am_key} # AM
    )
    (?P<am_middle>
        (
            ({letter})|
            ({digit})
        )
    )
    (?:
        (
            [\,\s,\-\&]|
            and
        )*
    )?
"""

optional = rf"""
    (?P<am_init_optional>
        {am_num}
    )?
    (?P<am_middle_optional>
        {letter}|
        {digit}
    )?
    (?:
        (
            [\,\s,\-\&]|
            and
        )*
    )?
"""

am_phrases = rf"""
    (?P<am_phrase>
        ({required})
        ({optional}){{1,3}}
    )
"""


constructed_am = CitationConstructor(
    label=DocketCategory.AM.value,
    short_category=DocketCategory.AM.name,
    group_name="am_phrase",
    init_name="am_init",
    docket_regex=am_phrases,
    key_regex=am_key,
    num_regex=Num.AM.allowed,
)


class CitationAM(DocketReportCitation):
    ...

    @classmethod
    def search(cls, text: str) -> Iterator[Self]:
        """Get all dockets matching the `AM` docket pattern, inclusive of their optional Report object.

        Examples:
            >>> text = "A.M. No. P-88-198, February 25, 1992, 206 SCRA 491."
            >>> cite = next(CitationAM.search(text))
            >>> cite.model_dump(exclude_none=True)
            {'publisher': 'SCRA', 'volume': '206', 'page': '491', 'context': 'A.M. No. P-88-198', 'category': 'AM', 'ids': 'P-88-198', 'docket_date': datetime.date(1992, 2, 25)}

        Args:
            text (str): Text to look for citation objects

        Yields:
            Iterator[Self]: Combination of Docket and Report pydantic model.
        """  # noqa E501
        for result in constructed_am.detect(text):
            yield cls(**result)



from collections.abc import Iterator
from typing import Self

from .models import CitationConstructor, DocketCategory, DocketReportCitation
from .models.misc import NUMBER_KEYWORD

separator = r"[\.\s]*"
digit = r"(\d{2}|P)[\d-]*"  # e.g. P-22-069

end_digit = r"""
    (\d{2,})|
    RTJ| # 21-105-RTJ
    MTJ| # JIB FPI No. 21-018-MTJ. August 17, 2022
    P| # 21-047-P
    SC| # 21-002-SC
    CA-J| # 21-003-CA-J
    METC # 	21-018-MTJ
"""

full_digit = rf"""
    {digit}
    ({end_digit})
"""

jib_key = rf"""
    (
        (
            j
            {separator}
            i
            {separator}
            b
            {separator}
        )
        (
            (FPI)|(F{separator}P{separator}I{separator})
        )?
        \s*
        {NUMBER_KEYWORD}
    )
"""

required = rf"""
    (?P<jib_init>
        {jib_key}
    )
    (?P<jib_middle>
        {full_digit}
    )
    (?:
        {separator}
    )?
"""


jib_phrases = rf"""
    (?P<jib_phrase>
        {required}
        [\,\s]*
    )
"""

constructed_jib = CitationConstructor(
    label=DocketCategory.JIB.value,
    short_category=DocketCategory.JIB.name,
    group_name="jib_phrase",
    init_name="jib_init",
    docket_regex=jib_phrases,
    key_regex=jib_key,
    num_regex=NUMBER_KEYWORD,
)


class CitationJIB(DocketReportCitation):
    ...

    @classmethod
    def search(cls, text: str) -> Iterator[Self]:
        """Get all dockets matching the `JIB` docket pattern, inclusive of their optional Report object.

        Examples:
            >>> text = "JIB FPI No. 21-018-MTJ. August 17, 2022"
            >>> cite = next(CitationJIB.search(text))
            >>> cite.model_dump(exclude_none=True)
            {'context': 'JIB FPI No. 21-018-MTJ.', 'category': 'JIB', 'ids': '21-018-MTJ', 'docket_date': datetime.date(2022, 8, 17)}

        Args:
            text (str): Text to look for citation objects

        Yields:
            Iterator[Self]: Combination of Docket and Report pydantic model.
        """  # noqa E501
        for result in constructed_jib.detect(text):
            yield cls(**result)
